import 'package:cresce_client/http_requests/formaters/decoders.dart';
import 'package:equatable/equatable.dart';
import 'package:http/http.dart' as http;

class HttpPost extends HttpRequest {
  HttpPost(
    String authority,
    Formatters formatters, {
    HttpResultHandler resultHandler,
  }) : super(authority, formatters, resultHandler: resultHandler);

  http.Client makeClient() => super.makeClient();

  Future<HttpResult> post(String url, [Serializable body]) async {
    return send(
      url,
      (url, headers, client) {
        return client.post(url, headers: headers, body: makeBody(body));
      },
    );
  }

  Object makeBody(Serializable body) => body?.serialize(formatters.encoder);
}

class HttpGet extends HttpRequest {
  HttpGet(
    String authority,
    Formatters formatters, {
    HttpResultHandler resultHandler,
  }) : super(authority, formatters, resultHandler: resultHandler);

  @override
  http.Client makeClient() => super.makeClient();

  Future<HttpResult> get(String url) {
    return send(
      url,
      (url, headers, client) {
        return client.get(url, headers: headers);
      },
    );
  }
}

typedef Future<http.Response> HttpSend(
  String url,
  Map<String, String> headers,
  http.Client client,
);

class HttpRequest {
  final String authority;
  final Formatters formatters;
  final HttpResultHandler resultHandler;

  HttpRequest(
    this.authority,
    this.formatters, {
    HttpResultHandler resultHandler,
  }) : resultHandler = resultHandler ?? HttpResultHandler(formatters);

  http.Client makeClient() => http.Client();

  Future<HttpResult> send(String url, HttpSend request) async {
    var client = makeClient();
    try {
      var response = await request(
        _makeUrl(url),
        _makeHeaders(),
        client,
      );

      return resultHandler.handle(response);
    } finally {
      client.close();
    }
  }

  String _makeUrl(String url) => _combine([authority, url]);

  Map<String, String> _makeHeaders() {
    return {
      'content-type': formatters.getContentType(),
    };
  }

  String _combine(List<String> list) {
    return list.map((part) {
      return part.endsWith('/') ? part.substring(0, part.length - 1) : part;
    }).join('/');
  }
}

class HttpResultHandler {
  final Formatters formatters;

  const HttpResultHandler(this.formatters);

  HttpResult handle(http.Response response) {
    return HttpResult(
      formatters.decoder,
      statusCode: response.statusCode,
      body: response.body,
    );
  }
}

class HttpResult extends Equatable {
  final int _statusCode;
  final String body;
  final Decoder decoder;

  HttpResult(
    this.decoder, {
    int statusCode,
    this.body,
  }) : _statusCode = statusCode;

  bool wasSuccess() => [200, 201].contains(_statusCode);

  T deserialize<T>(Deserialize obj) => cast<T>(
      shouldDeserialize() ? obj.deserialize(decoder.decode(body)) : obj);

  bool shouldDeserialize() => wasSuccess() && body != null;

  T cast<T>(Object obj) => obj as T;

  @override
  List<Object> get props => [_statusCode, body];

  List<T> deserializeList<T>(Deserialize obj) {
    var mapped = decoder.decode(body);

    final casting = mapped.cast<Map<String, dynamic>>();

    return casting.map<T>((f) {
      return obj.deserialize(f) as T;
    }).toList();
  }
}

abstract class Serializable {
  String serialize(Encoder encoder);
}

abstract class Deserialize {
  Object deserialize(Map<String, dynamic> data);
}

abstract class DeserializeList {
  List<Object> deserialize(List data);
}
